/*
 *  sample C  -- symbol table definition and manipulation
 */

#include "symbol.h"
#include "message.c"
#include "sc.tab.h"
#include <stdlib.h> 

/*
 *  Symbol table
 */

WORDPTR global = NULL,             /* global end of chain */
        local = NULL;              /* local end of chain */

/*
 * Block Table
 */

int blockLevel = 0 ;  /* current static block depth */

/*
 *  strsave
 */
char * strsave( char *s )
{
  char * cp = (char *) malloc( strlen(s)+1);

  if (cp)
  {
    strcpy(cp, s);
    return cp;
  }
  error("No more room to save strings");
  return NULL;
}/* strsave */

/*
 *   Add a new name to local region
 */

/*
 *  local points to a blind element, wich never changes, and after which
 *  new entries are added. Along the chain, at the bottom of the stack,
 *  are essencially never popped.
 *  The function create() will add a new, local entry to the symbol
 *  table for the name given as an argument.
 *  The new entry is marked as being undeclared.
 */

WORDPTR create( char * name )
{
  WORDPTR newEntry;

  newEntry = (WORDPTR) malloc( sizeof(WORD));

  if ( newEntry )
  {
    newEntry->next = local;
    local = newEntry;
    newEntry->name = strsave(name);
    newEntry->kind = UDEC;
    newEntry->blockLevel = 0;
    newEntry->S.num = NOTSET;

    return newEntry;
  }
  fatal( "No more room for symbols." );
} /* create */


/*
 *  move an entry from local to global region
 */


/*
 * if a function description is moved to the global end of the chain,
 * it is moved to follow "global" and "global" is then pointed to it.
 * This is acomplised by this routine, which has an argument a pointer
 * to the entry to be moved
 */

void move( WORDPTR symbol)
{

  if (symbol != NULL && local != NULL && global != NULL){
  /* unlink it from its present position */
  local = local->next;

  /* relink at global end of symtab */
  global->next = symbol;
  symbol->next = NULL;
  global = symbol;
  }
} /* move */


/*
 *  locate entry by name
 */

WORDPTR find( char * name)
{
  WORDPTR ptr;

  /* search symtab until match or end of symtab chain */
  ptr = local; 
  while (ptr)
  { 
    if ( ! ptr->name ) bug("Name not find");
    else
      /* return ptr if names match */
      if ( strcmp(ptr->name,name) == 0 ) return ptr;
    ptr = ptr->next;
  }

  /* search fails, return NULL */
  return (WORDPTR) 0;

}/* find */


/*
 *  interface for lexical analyzer:
 *  locate or enter Identifier, save text of Constant
 */

WORDPTR lookUp( char *s )
{
  WORDPTR p;

  p = find(s);

  if ( p ) return p;
  return create(s);
} /* lookUp */


/* 
 *  mark entry as part of parameterList
 */

WORDPTR linkParam( WORDPTR symbol, WORDPTR next)
{
  switch( symbol->kind )
  {
    case PARAM : error("duplicate parameter %s", symbol->name);
                 return next;
    case FUNC  :
    case UFUNC :
    case VAR   : symbol = create( symbol->name);   
    case UDEC  : break;
    default    : bug("linkParam");
  } /* switch ...*/

  symbol->kind = PARAM;
  symbol->blockLevel = blockLevel;
  symbol->S.link = next;
  return symbol;

} /* linkParam */



/*
 *  declare parameter
 */

WORDPTR makeParam( WORDPTR symbol)
{
  switch( symbol->kind )
  {
    case VAR   : if (symbol->blockLevel == 2) {
                   error("parameter %s declare twice", symbol->name);
                   return symbol;
                 } 
    case UDEC  : 
    case FUNC  :
    case UFUNC : error(" %s is not a parameter", symbol->name);
                 symbol = create(symbol->name);
    case PARAM : break;
    default    : bug("makeParam");
  } /* switch ...*/

  symbol->kind = VAR;
  symbol->blockLevel = blockLevel;
  return symbol;

} /* makeParam */

/*
 *  define a variable
 */

WORDPTR makeVar( WORDPTR symbol)
{
  switch( symbol->kind )
  {
    case VAR   :
    case FUNC  :
    case UFUNC : if (symbol->blockLevel == blockLevel
                     || symbol->blockLevel == 2 && blockLevel == 3)
                   error("duplicate name %s", symbol->name);
                 symbol = create(symbol->name);
    case UDEC  : break;
    case PARAM : error(" unexpected parameter %s", symbol->name);
                 break;
    default    : bug("makeVar");
  } /* switch ...*/

  symbol->kind = VAR;
  symbol->blockLevel = blockLevel;
  return symbol;

} /* makeVar */


/*
 * define a function
 */

WORDPTR makeFunc( WORDPTR symbol)
{
  switch( symbol->kind )
  {
    case UFUNC :
    case UDEC  : break;
    case VAR   : error("function name %s same as global variable", symbol->name);
                 return symbol;
    case FUNC  : error("duplicate function definition %s", symbol->name);
                 return symbol;
    default    : bug("makeFunc");
  } /* switch ...*/

  symbol->kind = FUNC;
  symbol->blockLevel = 1;
  
  return symbol;

} /* makeFunc */


/*
 * set of verify number of parameters
 */

void checkParam( WORDPTR symbol, int count)
{
  if ( symbol->S.num == NOTSET) symbol->S.num = count;
  else if ( symbol->S.num != count)
    warning( "function %s should have %d argument(s)", symbol->name, symbol->S.num);

} /* checkParam */


/*
 * default undeclared parameters, count
 */
int paramDefault( WORDPTR symbol)
{
  int count = 0;
  while (symbol)
  {
    ++count;
    if ( symbol->kind == PARAM ) symbol->kind = VAR;
    symbol = symbol->S.link;
  }
  return count;

} /* ParamDefault*/


/*
 * check reference or assignment to variable
 */

void checkVar( WORDPTR symbol)
{
  switch( symbol->kind )
  {
    case UDEC  : error( "undeclared variable %s", symbol->name);
                 break;
    case PARAM : error( "unexpected parameter %s", symbol->name);
                 break;
    case UFUNC :
    case FUNC  : error( "function %s used as variable", symbol->name);
    case VAR   : return ;
    default    : bug("checkVar");
  } /* switch ...*/

  symbol->kind = VAR;
  symbol->blockLevel = blockLevel;
  
} /* checkVar */


/*
 * check reference to function, implicitly declare it 
 */

void checkFunc( WORDPTR symbol)
{
  switch( symbol->kind )
  {
    case UDEC  : break;
    case PARAM : error( "unexpected parameter %s", symbol->name);
                 symbol->S.num = NOTSET;
                 return;
    case VAR   : error( "variable %s used as function", symbol->name);
                 symbol->S.num = NOTSET;
    case UFUNC :
    case FUNC  : return;    
    default    : bug("checkFunc");
  } /* switch ...*/

  move( symbol);
  symbol->kind = UFUNC;
  symbol->blockLevel = 1;
  
} /* checkFunc */



/*
 *    push block table
 */

void blockPush()
{
  ++blockLevel;

} /* blockPush */


/*
 *    pop block table
 */

void blockPop()
{
  WORDPTR ptr;
  
  ptr = local;
  if (ptr != NULL) {
    while ((ptr->blockLevel >= blockLevel) || (ptr->blockLevel == 0) )
    {
      if (! ptr->name) bug( "blockPop null name");

      if (ptr->kind == UFUNC) error("undefined function %s", ptr->name);
      free( ptr->name);
    
      local = ptr->next;
      free( ptr);

      ptr = local;
      if (ptr == NULL) break;
    } /* while ... */
  } 

  -- blockLevel;

} /* blockPop */


/*
 * initialize symbol and block table
 */

void init()
{
   blockPush();

   global = create( "main");
   global->kind = UFUNC;

} /* init */
